{-# LANGUAGE TemplateHaskell #-}

module ISets where

import qualified Data.List       as L
import           Test.QuickCheck

newtype ISet = ISet [(Int,Int)]
  deriving (Eq,Show)

valid :: ISet -> Bool
valid (ISet xys) =
  all (uncurry (<=)) xys &&
  and [y+1<x | ((_,y),(x,_)) <- zip xys (drop 1 xys)]

-- toList --------------------------------------------------------------

toList :: ISet -> [Int]
toList (ISet xys) =
  concat [[x..y] | (x,y) <- xys]

-- Generator for ISets -------------------------------------------------

instance Arbitrary ISet where
  arbitrary = do Ordered xs <- arbitrary
                 return $ ISet $ make xs
    where make (x:y:xs) = (x,y):make (map (+2) xs)
          make _        = []
  shrink (ISet xs) = filter valid $ map ISet $ shrink xs

-- member --------------------------------------------------------------

prop_Member x s =
  (x `member` s) === (x `elem` toList s)

prop_MemberTest1 = prop_Member 0 (ISet [])
prop_MemberTest2 = prop_Member 0 (ISet [(0,0)])
prop_MemberTest3 = prop_Member 0 (ISet [(-1,0)])
prop_MemberTest4 = prop_Member 1 (ISet [(0,0)])
prop_MemberTest5 = prop_Member 2 (ISet [(0,0),(2,2)])

member x (ISet xys) = member' x xys

member' x [] = False
member' x ((x',y'):xys) =
  (x >= x' && x <= y') || member' x xys

-- insert --------------------------------------------------------------

prop_InsertValid x s = valid (insert x s)

prop_InsertModel x s =
  toList (insert x s) === L.insert x (toList s L.\\ [x])

insert :: Int -> ISet -> ISet
insert x (ISet xys) = ISet (insert' x xys)

insert' x xys = undefined

-- delete --------------------------------------------------------------

prop_DeleteValid x s = valid (delete x s)

prop_DeleteModel x s =
  toList (delete x s) === L.delete x (toList s)

delete x (ISet xys) = ISet (delete' x xys)

delete' x xys = undefined

-- union ---------------------------------------------------------------

prop_UnionValid s s' = valid (union s s')

prop_UnionModel s s' =
  toList (union s s') === L.sort (L.union (toList s) (toList s'))

union (ISet xys) (ISet xys') = ISet (union' xys xys')

union' = undefined

-- Enable running all properties ---------------------------------------
return []

runTests :: IO Bool
runTests = $quickCheckAll
